/*
 * Copyright (C) 1999, David Luyer and The University of Western Australia
 *
 * This file is released under the GNU General Public License.
 * Please read the file "COPYING" for license information.
 * If you did not receive the file "COPYING" with this file, a copy is
 * available by e-mail from David Luyer <luyer@ucs.uwa.edu.au>.
 */
#define _GNU_SOURCE 1

#include <dlfcn.h>
#include <fcntl.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/stat.h>

int keep_running = 1;
char *shm_ptr = NULL;
char input_char = 0;

void **user_data = NULL;
int user_data_version = 0;

static int get_id(char *filename) {
  int fd;
  char buf[64];

  if((fd = open(filename, O_RDONLY)) == -1) {
    perror("open");
    exit(1);
  }
  if(read(fd, buf, 64) == -1) {
    perror("open");
    exit(1);
  }
  close(fd);
  return atoi(buf);
}

static void input_mode(int on) {
static struct termios saved_settings;
struct termios settings;

  if(!on || on > 1) {
    tcsetattr(0, TCSANOW, &saved_settings);
    if(on) {
      shmdt(shm_ptr);
      exit(1);
    }
    return;
  }
  if(tcgetattr(0, &saved_settings) < 0) {
    perror("tcgetatter");
    exit(1);
  }
  settings = saved_settings;
  settings.c_lflag &= ~(ICANON|ISIG|ECHO);
  if(tcsetattr(0, TCSANOW, &settings) < 0) {
    perror("tcgetatter");
    exit(1);
  }
}

#define LOAD_LIBRARY \
  library_mtime = sb.st_mtime; \
  dlptr = dlopen(argv[2], RTLD_NOW); \
  if(!dlptr) { \
    fprintf(stderr, "%s\n", dlerror()); \
    exit(1); \
  } \
  process_input = (void (*)(void)) dlsym(dlptr, "process_input"); \
  if((loadlibrary_str = dlerror()) != NULL) { \
    fprintf(stderr, "%s\n", loadlibrary_str); \
    exit(1); \
  }

extern int main(int argc, char **argv) {
  int current_id;
  void *dlptr;
  time_t library_mtime;
  struct stat sb;
  char *loadlibrary_str;

  void (* process_input) (void);

  if(argc != 3) {
    printf(
"Usage:\n"
"\tshm_client shmid_filename shared_object.so\n"
"\n"
"Operation:\n"
"\tThis program monitors a shared memory space maintained by shm_daemon,\n"
"\tgetting the initial shmid from shmid_filename.  What this client does\n"
"\tis determined by the process_input() function in the shared object file\n"
"\tcalled with as a second argument.  This function is called after each\n"
"\tcharacter of input.  The shared object file should access the variables\n"
"\tkeep_running (int), shm_ptr (char *) and input_char (char) from this\n"
"\tmain executable and may store data in user_data (void **) and\n"
"\tuser_data_version (int).\n\n"
"\tThe shared object file is also monitored for changes and a new copy is\n"
"\tloaded if it is updated.  The shared object file must not change its\n"
"\tcurrent directory if it wants to be able to be loaded as ./file.so.\n"
"Copyright:\n"
"\tCopyright (c) 1999, David Luyer and The University of Western Australia\n"
"\tThis program may comes with ABSOLUTELY NO WARRANTY and may be\n"
"\tdistributed as detailed in the file COPYING.\n"
          );
    exit(1);
  }
  current_id = get_id(argv[1]);
  if(current_id == 0) {
    fprintf(stderr, "WARNING: shm identifier of 0 is from shmid_file %s.\n", argv[1]);
  }
  if(argv[2][0] != '/' && argv[2][0] != '.') {
    fprintf(stderr, "The shared object filename must be an anchored pathname.\n");
    exit(1);
  }
  if((shm_ptr = shmat(current_id, 0, SHM_RDONLY)) == (char *)-1) {
    perror("shmat");
    exit(1);
  }
  if(stat(argv[2], &sb) == -1) {
    perror(argv[2]);
    exit(1);
  }
  LOAD_LIBRARY;
  signal(SIGBUS, input_mode);
  signal(SIGTERM, input_mode);
  signal(SIGSEGV, input_mode);
  input_mode(1);
  for (;;) {
    process_input();
    if(!keep_running)
      break;
    read(0, &input_char, 1);
    if(stat(argv[2], &sb) != -1 && sb.st_mtime != library_mtime) {
      dlclose(dlptr);
      input_mode(0);
      LOAD_LIBRARY;
      input_mode(1);
    }
    if(*(int *)shm_ptr != -1) {
      current_id = *(int *)shm_ptr;
      shmdt(shm_ptr);
      shm_ptr = shmat(current_id, 0, SHM_RDONLY);
      if(shm_ptr == (char *)-1) {
        input_mode(0);
        current_id = get_id(argv[1]);
        if((shm_ptr = shmat(current_id, 0, SHM_RDONLY)) == (char *)-1) {
          perror("shmat");
          exit(1);
        }
        input_mode(1);
      }
    }
  }
  input_mode(0);
  shmdt(shm_ptr);
  return 0;
}
