/*
 * Copyright (C) 1999, David Luyer and The University of Western Australia
 *
 * This file is released under the GNU General Public License.
 * Please read the file "COPYING" for license information.
 * If you did not receive the file "COPYING" with this file, a copy is
 * available by e-mail from David Luyer <luyer@ucs.uwa.edu.au>.
 */
#define _GNU_SOURCE 1

#include <time.h>
#include <fcntl.h>
#include <stdio.h>
#include <signal.h>
#include <malloc.h>
#include <limits.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <sys/types.h>

void copy_file(int fd, char *where, off_t size) {
  while(size > SSIZE_MAX) {
    if(read(fd, where, SSIZE_MAX) != SSIZE_MAX) {
      perror("read");
      exit(1);
    }
    where += SSIZE_MAX;
    size -= SSIZE_MAX;
  }
  if(read(fd, where, size) != size) {
    perror("read");
    exit(1);
  }
}

int updated(time_t date, char *filename) {
static time_t last;
struct stat sb;

  if(date) {
    last = date;
    return 0;
  }
  if(stat(filename, &sb) == -1) {
    return 0;
  }
  if(sb.st_mtime != last) {
    return 1;
  }
  return 0;
}

int get_new_shm(char *file) {
static int last_id = 0, warned = 0;
static char *optr;
int fd, shmid;
char *ptr;
key_t try;
struct stat sb;

  fd = open(file, O_RDONLY);
  if(fd == -1) {
    perror("open");
    exit(1);
  }
  if(fstat(fd, &sb) == -1) {
    perror("fstat");
    exit(1);
  }
  try = time(NULL);
  for(;;) {
    if((shmid = shmget(try, sb.st_size + sizeof(off_t) + sizeof(int) + 1,
                     IPC_CREAT|IPC_EXCL|0644)) != -1)
      break;
    try++;
  }
  if((ptr = shmat(shmid, 0, 0)) == (char *)-1) {
    perror("shmat");
    shmctl(shmid, IPC_RMID, 0);
    exit(1);
  }
  if(shmctl(shmid, SHM_LOCK, 0) != 0 && !warned) {
    warned = 1;
    perror("shmctl");
    fprintf(stderr, "WARNING: Continuing without locking down memory pages.\n");
  }
  *(int *)ptr = -1;
  ptr += sizeof(int);
  *(off_t *)ptr = sb.st_size;
  copy_file(fd, ptr + sizeof(off_t), sb.st_size);
  ptr[sizeof(off_t) + sb.st_size] = '\0';
  close(fd);
  updated(sb.st_mtime, NULL);
  if(optr) {
    *(int *)optr = shmid;
    shmdt(optr);
    shmctl(last_id, IPC_RMID, 0);
  }
  optr = ptr - sizeof(int);
  last_id = shmid;
  return shmid;
}

void note_id(int id, char *filename) {
  char buf[32];
  int i, fd;

  i = umask(022);
  if((fd = open(filename, O_CREAT|O_TRUNC|O_WRONLY, 0644)) == -1) {
    perror("open");
    exit(1);
  }
  sprintf(buf, "%d\n", id);
  if(write(fd, buf, strlen(buf)) == -1) {
    perror("write");
    exit(1);
  }
  umask(i);
}

struct int_list {
  int val;
  struct int_list *next;
};

volatile int got_signal = 0;

void signal_fn(int sig) {
  got_signal = 1;
}

int main(int argc, char **argv) {
  int current_id;

  if (argc != 3) {
    printf(
"Usage:\n"
"\tshm_daemon data_filename shmid_filename\n"
"\n"
"Operation:\n"
"\tThis program maintains a copy of data_filename in a shared memory\n"
"\tsegment.  The current shared memory segment number is stored in the\n"
"\tfile smhid_filename.  Whenever the file is updated, a new segment is\n"
"\tcreated with the new file.  The old segment is deleted, but will stay\n"
"\taround as long as other processes are attached to it.\n\n"
"\tThe shared memory segment contains the new shared memory smhid (or -1,\n"
"\trepresenting none), the file length and the file.\n"
"\n"
"Copyright:\n"
"\tCopyright (c) 1999, David Luyer and The University of Western Australia\n"
"\tThis program may comes with ABSOLUTELY NO WARRANTY and may be\n"
"\tdistributed as detailed in the file COPYING.\n"
          );
    exit(1);
  }
  signal(SIGINT, signal_fn);
  signal(SIGBUS, signal_fn);
  signal(SIGSEGV, signal_fn);
  signal(SIGQUIT, signal_fn);
  signal(SIGTERM, signal_fn);
  current_id = get_new_shm(argv[1]);
  note_id(current_id, argv[2]);
  if(fork())
    exit(0);
  close(0);
  close(1);
  close(2);
  open("/dev/null", O_RDWR);
  dup2(0, 1);
  dup2(0, 2);
  setsid();
  for(;;) {
    sleep(60);
    if (got_signal)
      break;
    if (updated(0, argv[1])) {
      current_id = get_new_shm(argv[1]);
      note_id(current_id, argv[2]);
    }
  }
  shmctl(current_id, IPC_RMID, 0);
  return 0;
}
